import java.io.*;
import java.net.*;
import java.sql.*;

public class JSQLExec {
    private static String jdbcDriverPathOrJarFileName = null;
    private static String jdbcDriverClassName = null;
    private static String jdbcUserName = null;
    private static String jdbcPassword = null;
    private static String jdbcURL = null;
    private static Connection connection = null;
    private static Statement statement = null;
    private static ResultSet resultSet = null;
    private static boolean isUpdate = false;
    private static String inputFileName = null;
    private static String outputFileName = null;
    private static OutputStreamWriter outputFileWriter = null;
    private static String nextExceptionMessage = null;
    private static String lineSeparator = null;

    public static void main( String[] args ) {
        lineSeparator = System.getProperty( "line.separator" );

        try {
            guardedMain( args );
        } catch ( Exception e ) {
            Exception ee;
            if ( nextExceptionMessage == null ) {
                ee = new Exception( "No error message available", e );
            } else {
                ee = new Exception( nextExceptionMessage, e );
            }
            logExceptions( ee );
            cleanAndExit( 1 );
        }
        cleanAndExit( 0 );
    }

    private static void guardedMain( String[] args ) throws Exception {
        if ( !parseArguments( args ) ) {
            printUsage();
            cleanAndExit( 1 );
        }

        ClassLoader classLoader = ClassLoader.getSystemClassLoader();
        if ( jdbcDriverPathOrJarFileName.length() != 0 ) {
            classLoader = new URLClassLoader( new URL[] {new File( jdbcDriverPathOrJarFileName ).toURL()} );
        }

        String sqlCommand = readSQLCommand( inputFileName );

        outputFileWriter = openFileForWritingUTF8( outputFileName );

        pushNextExceptionMessage( "Error occured while loading JDBC driver class: " + jdbcDriverClassName );
        classLoader.loadClass( jdbcDriverClassName );
        popExceptionMessage();

        pushNextExceptionMessage( "Error occured while loading JDBC driver class: " + jdbcDriverClassName );
        Class.forName( jdbcDriverClassName );
        popExceptionMessage();

        pushNextExceptionMessage( "Error occured while creating JDBC connection" );
        connection = DriverManager.getConnection( jdbcURL, jdbcUserName, jdbcPassword );
        popExceptionMessage();

        pushNextExceptionMessage( "Error occured while executing SQL command: " + sqlCommand );
        execSQL( connection, sqlCommand, isUpdate );
        popExceptionMessage();
    }

    private static void execSQL( Connection connection, String sqlString, boolean isUpdate ) throws Exception {
        statement = connection.createStatement();
        if ( isUpdate ) {
            int updateCount = statement.executeUpdate( sqlString );
            outputUpdateCount( updateCount );
        } else {
            resultSet = statement.executeQuery( sqlString );
            outputResultSet( resultSet );
            resultSet.close();
        };
        statement.close();
        statement = null;
    }

    private static void outputUpdateCount( int updateCount ) throws Exception {
        output( "UpdateCount" );
        output( lineSeparator );
        output( "" + updateCount );
        output( lineSeparator );
    }

    private static void output( String s ) throws Exception {
        //System.out.print( s );
        outputFileWriter.write( s );
    }

    private static void outputResultSet( ResultSet resultSet ) throws Exception {
        // TODO
        ResultSetMetaData m = resultSet.getMetaData();
        int columnCount = m.getColumnCount();

        for ( int col = 1; col <= columnCount; col++ ) {
            String l = m.getColumnLabel( col );
            output( l );
            if ( col < columnCount ) {
                output( "," );
            }
        }
        output( lineSeparator );

        while ( resultSet.next() ) {
            for ( int col = 1; col <= columnCount; col++ ) {
                String escapedValue = toEscapedCSVString( resultSet.getString( col ) );
                output( escapedValue );
                if ( col < columnCount ) {
                    output( "," );
                }
            }
            output( lineSeparator );
        }
    }

    private static String toEscapedCSVString( String s ) {
        if ( s.indexOf( "," ) == -1 ) {
            return s;
        }
        s = s.replaceAll( "\"", "\"\"" );
        return "\"" + s + "\"";
    }

    private static void outputItem( String value ) throws Exception {
        pushNextExceptionMessage( "Error while outputting value: " + value );
        outputFileWriter.write( value );
        outputFileWriter.write( lineSeparator );
        popExceptionMessage();
    }

    private static String readSQLCommand( String inputFileName ) throws Exception {
        pushNextExceptionMessage( "Error occured while reading SQL commands from file: " + inputFileName );
        String s = readFileUTF8( inputFileName );
        popExceptionMessage();
        return s;
    }

    private static void pushNextExceptionMessage( String message ) {
        // TODO - Implement push/pop as a list
        nextExceptionMessage = message;
    }

    private static void popExceptionMessage() {
        // TODO - Implement push/pop as a list
        nextExceptionMessage = null;
    }

    private static OutputStreamWriter openFileForWritingUTF8( String fileName ) throws Exception {
        OutputStreamWriter w = new OutputStreamWriter( 
                                    new BufferedOutputStream( 
                                        new FileOutputStream( fileName )
                                     ),
                                    "UTF-8" );
        return w;
    }

    private static InputStreamReader openFileForReadingUTF8( String fileName ) throws Exception {
        InputStreamReader w = new InputStreamReader( 
                                    new BufferedInputStream( 
                                        new FileInputStream( fileName )
                                     ),
                                    "UTF-8" );
        return w;
    }

    private static void writeFileUTF8( String fileName, String fileContents ) throws Exception {
        OutputStreamWriter w = openFileForWritingUTF8( fileName );
        w.write( fileContents );
        w.close();
    }

    private static String readFileUTF8( String fileName ) throws Exception {
        InputStreamReader r = openFileForReadingUTF8( fileName );
        StringBuffer sb = new StringBuffer();
        int bufferSize = 1024;
        char[] buffer = new char[ bufferSize ];
        while ( true ) {
            int read = r.read( buffer, 0, bufferSize );
            if ( read == -1 ) {
                break;
            }
            sb.append( buffer, 0, read );
        }
        return sb.toString();
    }

    private static void logExceptions( Exception e ) {
        // TODO - Actually log to log file?
        e.printStackTrace();
    }

    private static boolean parseArguments( String[] args ) throws Exception {
        if ( args.length == 1 ) {
            String s = readFileUTF8( args[ 0 ] );
            s = s.replaceAll( "\r\n", "\n" );
            s = s.replaceAll( "\r", "\n" );
            args = s.split( "\n" );
        }

        if ( args.length == 8 && ( args[ 0 ].equals( "select" ) || args[ 0 ].equals( "update" ) ) ) {
            int i = 0;
            isUpdate = args[ i++ ].equals( "update" );
            jdbcDriverPathOrJarFileName = args[ i++ ];
            jdbcDriverClassName = args[ i++ ];
            jdbcURL = args[ i++ ];
            jdbcUserName = args[ i++ ];
            jdbcPassword = args[ i++ ];
            inputFileName = args[ i++ ];
            outputFileName = args[ i++ ];
            return true;
        }

        return false;
    }

    private static void cleanAndExit( int n ) {
        if ( outputFileWriter != null ) {
            try {
                outputFileWriter.close();
            } catch ( Exception e ) {
                String message = "Error occured while closing output file upon exit: " + outputFileName;
                Exception ee = new Exception( message, e );
                logExceptions( ee );
            }
        }

        if ( statement != null ) {
            try {
                statement.close();
            } catch ( Exception e ) {
                String message = "Error occured while closing JDBC statement upon exit";
                Exception ee = new Exception( message, e );
                logExceptions( ee );
            }
        }

        if ( connection != null ) {
            try {
                connection.close();
            } catch ( Exception e ) {
                String message = "Error occured while closing JDBC connection upon exit";
                Exception ee = new Exception( message, e );
                logExceptions( ee );
            }
        }

        System.exit( n );
    }

    private static void printUsage() {
        System.out.println( "USAGE:" );
        System.out.println();
        System.out.println( "  JSQLExec [ select | update ] <jdbc_driver_path_or_jar_file_name> <jdbc_driver_class_name> <jdbc_url> <jdbc_user_name> <jdbc_password> <input_file_name> <csv_output_file_name>" );
        System.out.println();
        System.out.println( "  JSQLExec <parameter_file_name>" );
        System.out.println();
        System.out.println( "      Above parameters will be read from the specified file." );
        System.out.println( "      Each parameter must be on a separate line." );
        System.out.println( "      The parameters must not be quoted." );
        System.out.println( "      The parameter file must be UTF-8 encoded." );
        System.out.println();
        System.out.println( "EXAMPLES:" );
        System.out.println();
        System.out.println( "  JSQLExec select \"\" sun.jdbc.odbc.JdbcOdbcDriver \"jdbc:odbc:Driver={Microsoft Excel Driver (*.xls)};DBQ=excel_data.xls\" \"\" \"\" input_select.txt output.txt" );
        System.out.println();
        System.out.println( "  JSQLExec update \"\" sun.jdbc.odbc.JdbcOdbcDriver \"jdbc:odbc:Driver={Microsoft Excel Driver (*.xls)};DBQ=excel_data.xls;ReadOnly=false\" \"\" \"\" input_update.txt output.txt" );
        System.out.println();
        System.out.println( "  JSQLExec parameters.txt" );
    }
}
